@@ -23,10 +23,18 @@
 
 #include "xinput.h"
 #include <string.h>
-#include <X11/extensions/XIproto.h> /* for XI_Device***ChangedNotify */
+
+enum print_format {
+    FORMAT_NONE,
+    FORMAT_SHORT,
+    FORMAT_LONG,
+    FORMAT_NAME,
+    FORMAT_ID,
+};
+
 
 static void
-print_info(Display* dpy, XDeviceInfo	*info, Bool shortformat)
+print_info(Display* dpy, XDeviceInfo	*info, enum print_format format)
 {
     int			i,j;
     XAnyClassPtr	any;
@@ -35,6 +43,16 @@ print_info(Display* dpy, XDeviceInfo	*info, Bool shortformat)
     XValuatorInfoPtr	v;
     XAxisInfoPtr	a;
 
+    if (format == FORMAT_NAME)
+    {
+        printf("%s\n", info->name);
+        return;
+    } else if (format == FORMAT_ID)
+    {
+        printf("%ld\n", info->id);
+        return;
+    }
+
     printf("\"%s\"\tid=%ld\t[", info->name, info->id);
 
     switch (info->use) {
@@ -59,11 +77,14 @@ print_info(Display* dpy, XDeviceInfo	*info, Bool shortformat)
     }
     printf("]\n");
 
-    if (shortformat)
+    if (format == FORMAT_SHORT)
         return;
 
-    if(info->type != None)
-	printf("\tType is %s\n", XGetAtomName(dpy, info->type));
+    if (info->type != None) {
+	char *type = XGetAtomName(dpy, info->type);
+	printf("\tType is %s\n", type);
+	XFree(type);
+    }
 
     if (info->num_classes > 0) {
 	any = (XAnyClassPtr) (info->inputclassinfo);
@@ -104,47 +125,20 @@ print_info(Display* dpy, XDeviceInfo	*info, Bool shortformat)
 }
 
 static int list_xi1(Display     *display,
-                    int	        argc,
-                    char        *argv[],
-                    char        *name,
-                    char        *desc)
+                    enum print_format format)
 {
     XDeviceInfo		*info;
     int			loop;
-    int                 shortformat = False;
-    int                 daemon = False;
+    int                 num_devices;
 
-    shortformat = (argc == 1 && strcmp(argv[0], "--short") == 0);
-    daemon = (argc == 1 && strcmp(argv[0], "--loop") == 0);
-
-    if (argc == 0 || shortformat || daemon) {
-	int		num_devices;
-
-        do {
-            info = XListInputDevices(display, &num_devices);
-            for(loop=0; loop<num_devices; loop++) {
-                print_info(display, info+loop, shortformat);
-            }
-        } while(daemon);
-    } else {
-	int	ret = EXIT_SUCCESS;
-
-	for(loop=0; loop<argc; loop++) {
-	    info = find_device_info(display, argv[loop], False);
-
-	    if (!info) {
-		fprintf(stderr, "unable to find device %s\n", argv[loop]);
-		ret = EXIT_FAILURE;
-	    } else {
-		print_info(display, info, shortformat);
-	    }
-	}
-	return ret;
+    info = XListInputDevices(display, &num_devices);
+    for(loop=0; loop<num_devices; loop++) {
+        print_info(display, info+loop, format);
     }
     return EXIT_SUCCESS;
 }
 
-#ifdef HAVE_XI2
+#if HAVE_XI2
 /* also used from test_xi2.c */
 void
 print_classes_xi2(Display* display, XIAnyClassInfo **classes,
@@ -155,19 +149,23 @@ print_classes_xi2(Display* display, XIAnyClassInfo **classes,
     printf("\tReporting %d classes:\n", num_classes);
     for (i = 0; i < num_classes; i++)
     {
-        printf("\t\tClass originated from: %d\n", classes[i]->sourceid);
+        printf("\t\tClass originated from: %d. Type: ", classes[i]->sourceid);
         switch(classes[i]->type)
         {
             case XIButtonClass:
                 {
                     XIButtonClassInfo *b = (XIButtonClassInfo*)classes[i];
                     char *name;
+                    printf("XIButtonClass\n");
                     printf("\t\tButtons supported: %d\n", b->num_buttons);
                     printf("\t\tButton labels:");
                     for (j = 0; j < b->num_buttons; j++)
                     {
                         name = (b->labels[j]) ? XGetAtomName(display, b->labels[j]) : NULL;
-                        printf(" %s", (name) ? name : "None");
+                        if (name)
+                            printf(" \"%s\"", name);
+                        else
+                            printf(" None");
                         XFree(name);
                     }
                     printf("\n");
@@ -182,6 +180,7 @@ print_classes_xi2(Display* display, XIAnyClassInfo **classes,
             case XIKeyClass:
                 {
                     XIKeyClassInfo *k = (XIKeyClassInfo*)classes[i];
+                    printf("XIKeyClass\n");
                     printf("\t\tKeycodes supported: %d\n", k->num_keycodes);
                 }
                 break;
@@ -190,6 +189,10 @@ print_classes_xi2(Display* display, XIAnyClassInfo **classes,
                     XIValuatorClassInfo *v = (XIValuatorClassInfo*)classes[i];
                     char *name = v->label ?  XGetAtomName(display, v->label) : NULL;
 
+                    /* Bug in X servers 1.7..1.8.1, mode was | OutOfProximity */
+                    v->mode &= DeviceMode;
+
+                    printf("XIValuatorClass\n");
                     printf("\t\tDetail for Valuator %d:\n", v->number);
                     printf("\t\t  Label: %s\n",  (name) ? name : "None");
                     printf("\t\t  Range: %f - %f\n", v->min, v->max);
@@ -201,6 +204,41 @@ print_classes_xi2(Display* display, XIAnyClassInfo **classes,
                     XFree(name);
                 }
                 break;
+#if HAVE_XI21
+            case XIScrollClass:
+                {
+                    XIScrollClassInfo *s = (XIScrollClassInfo*)classes[i];
+
+                    printf("XIScrollClass\n");
+                    printf("\t\tScroll info for Valuator %d\n", s->number);
+                    printf("\t\t  type: %d (%s)\n", s->scroll_type,
+                           (s->scroll_type == XIScrollTypeHorizontal) ? "horizontal" :
+                              (s->scroll_type == XIScrollTypeVertical) ? "vertical" : "unknown");
+                    printf("\t\t  increment: %f\n", s->increment);
+                    printf("\t\t  flags: 0x%x", s->flags);
+                    if (s->flags) {
+                        printf(" (");
+                        if (s->flags & XIScrollFlagNoEmulation)
+                            printf(" no-emulation ");
+                        if (s->flags & XIScrollFlagPreferred)
+                            printf(" preferred ");
+                        printf(")");
+                    }
+                    printf("\n");
+                }
+                break;
+#endif
+#if HAVE_XI22
+            case XITouchClass:
+                {
+                    XITouchClassInfo *t = (XITouchClassInfo*)classes[i];
+
+                    printf("XITouchClass\n");
+                    printf("\t\tTouch mode: %s\n",
+                           (t->mode == XIDirectTouch) ? "direct" : "dependent");
+                    printf("\t\tMax number of touches: %d\n", t->num_touches);
+                }
+#endif
         }
     }
 
@@ -208,8 +246,18 @@ print_classes_xi2(Display* display, XIAnyClassInfo **classes,
 }
 
 static void
-print_info_xi2(Display* display, XIDeviceInfo *dev, Bool shortformat)
+print_info_xi2(Display* display, XIDeviceInfo *dev, enum print_format format)
 {
+    if (format == FORMAT_NAME)
+    {
+        printf("%s\n", dev->name);
+        return;
+    } else if (format == FORMAT_ID)
+    {
+        printf("%d\n", dev->deviceid);
+        return;
+    }
+
     printf("%-40s\tid=%d\t[", dev->name, dev->deviceid);
     switch(dev->use)
     {
@@ -230,7 +278,7 @@ print_info_xi2(Display* display, XIDeviceInfo *dev, Bool shortformat)
             break;
     }
 
-    if (shortformat)
+    if (format == FORMAT_SHORT)
         return;
 
     if (!dev->enabled)
@@ -240,28 +288,14 @@ print_info_xi2(Display* display, XIDeviceInfo *dev, Bool shortformat)
 }
 
 
-int
-list_xi2(Display	*display,
-         int	argc,
-         char	*argv[],
-         char	*name,
-         char	*desc)
+static int
+list_xi2(Display *display,
+         enum print_format format)
 {
-    int major = XI_2_Major,
-        minor = XI_2_Minor;
     int ndevices;
-    int i, j, shortformat;
+    int i, j;
     XIDeviceInfo *info, *dev;
 
-    shortformat = (argc == 1 && strcmp(argv[0], "--short") == 0);
-
-    if (XIQueryVersion(display, &major, &minor) != Success ||
-        (major * 1000 + minor) < (XI_2_Major * 1000 + XI_2_Minor))
-    {
-        fprintf(stderr, "XI2 not supported.\n");
-        return EXIT_FAILURE;
-    }
-
     info = XIQueryDevice(display, XIAllDevices, &ndevices);
 
     for(i = 0; i < ndevices; i++)
@@ -269,12 +303,15 @@ list_xi2(Display	*display,
         dev = &info[i];
         if (dev->use == XIMasterPointer || dev->use == XIMasterKeyboard)
         {
-            if (dev->use == XIMasterPointer)
-                printf("⎡ ");
-            else
-                printf("⎣ ");
+            if (format == FORMAT_SHORT || format == FORMAT_LONG)
+            {
+                if (dev->use == XIMasterPointer)
+                    printf("⎡ ");
+                else
+                    printf("⎣ ");
+            }
 
-            print_info_xi2(display, dev, shortformat);
+            print_info_xi2(display, dev, format);
             for (j = 0; j < ndevices; j++)
             {
                 XIDeviceInfo* sd = &info[j];
@@ -282,8 +319,9 @@ list_xi2(Display	*display,
                 if ((sd->use == XISlavePointer || sd->use == XISlaveKeyboard) &&
                      (sd->attachment == dev->deviceid))
                 {
-                    printf("%s   ↳ ", dev->use == XIMasterPointer ? "⎜" : " ");
-                    print_info_xi2(display, sd, shortformat);
+                    if (format == FORMAT_SHORT || format == FORMAT_LONG)
+                        printf("%s   ↳ ", dev->use == XIMasterPointer ? "⎜" : " ");
+                    print_info_xi2(display, sd, format);
                 }
             }
         }
@@ -295,7 +333,7 @@ list_xi2(Display	*display,
         if (dev->use == XIFloatingSlave)
         {
             printf("∼ ");
-            print_info_xi2(display, dev, shortformat);
+            print_info_xi2(display, dev, format);
         }
     }
 
@@ -312,11 +350,61 @@ list(Display	*display,
      char	*name,
      char	*desc)
 {
-#ifdef HAVE_XI2
-    if (xinput_version(display) == XI_2_Major)
-        return list_xi2(display, argc, argv, name, desc);
+    enum print_format format = FORMAT_NONE;
+    int arg_dev = 1;
+
+    if (argc >= 1)
+    {
+        if (strcmp(argv[0], "--short") == 0)
+            format = FORMAT_SHORT;
+        else if (strcmp(argv[0], "--long") == 0)
+            format = FORMAT_LONG;
+        else if (strcmp(argv[0], "--name-only") == 0)
+            format = FORMAT_NAME;
+        else if (strcmp(argv[0], "--id-only") == 0)
+            format = FORMAT_ID;
+        else
+            arg_dev--;
+    }
+
+    if (argc > arg_dev)
+    {
+        if (format == FORMAT_NONE)
+            format = FORMAT_LONG;
+#if HAVE_XI2
+        if (xinput_version(display) == XI_2_Major)
+        {
+            XIDeviceInfo *info = xi2_find_device_info(display, argv[arg_dev]);
+
+            if (!info) {
+                fprintf(stderr, "unable to find device %s\n", argv[arg_dev]);
+                return EXIT_FAILURE;
+            } else {
+                print_info_xi2(display, info, format);
+                return EXIT_SUCCESS;
+            }
+        } else
 #endif
-    return list_xi1(display, argc, argv, name, desc);
+        {
+            XDeviceInfo *info = find_device_info(display, argv[arg_dev], False);
+
+            if (!info) {
+                fprintf(stderr, "unable to find device %s\n", argv[arg_dev]);
+                return EXIT_FAILURE;
+            } else {
+                print_info(display, info, format);
+                return EXIT_SUCCESS;
+            }
+        }
+    } else {
+        if (format == FORMAT_NONE)
+            format = FORMAT_SHORT;
+#if HAVE_XI2
+        if (xinput_version(display) == XI_2_Major)
+            return list_xi2(display, format);
+#endif
+        return list_xi1(display, format);
+    }
 }
 
 /* end of list.c */
